#include <iostream>
#include <Cntm/Containers/HandledList.h>
#include <Cntm/Containers/Register.h>
#include <Cntm/Containers/RegisterEx.h>

using namespace Cntm;
using namespace std;

struct ValueType
{
	short i, j;
	
	ValueType(short I, short J): i(I), j(J)  { cerr << "{!V}"; }

	ValueType(const ValueType& Src): i(Src.i), j(Src.j) { cerr << "{+V}"; }
	
	~ValueType() { i = j = -1; cerr << "{-V}"; }
};

struct StoreType
{
	int k;
	
	StoreType(): k(0)  { cerr << "{@S}"; }
	StoreType(const StoreType& Src): k(Src.k) { cerr << "{+S}"; }
	StoreType& operator = (const StoreType& Src) { k = Src.k; cerr << "{=S}"; return *this; }
	~StoreType() { k = -1; cerr << "{-S}"; }

	//   ValueType
	StoreType(const ValueType& Src): k(Src.i * 65536 + Src.j)  { cerr << "{>S}"; }
	
	//   ValueType
	operator ValueType() { return ValueType(k / 65536, k % 65536); }
};

void ContainerHandledListTypecasting()
{
	//     
	cout << "    : ";
	typedef HandledList<ValueType, StoreType> Reg0;
	Reg0 r0;
	
	cout << endl << "     ValueType(2, 5) ";
	Reg0::Handle h1 = r0.Add(ValueType(2, 5));
	
	cout << endl << "      ";
	Reg0::Handle h2 = h1;	
	
	cout << endl << "          ";
	Reg0::Enumerator e1 = r0.GetEnumerator();
	++e1;

	cout << endl << "      ";
	Reg0::Enumerator e2 = e1;

	cout << endl << "    ";
	cout << " [1.1]=ValueType(" << e1->i << ", " << e1->j << ") ";
	++e1;
	cout << " [1.2]=ValueType(" << e1->i << ", " << e1->j << ") ";

	cout << endl << "    ";
	cout << " [2.1]=ValueType(" << (*e2).i << ", " << (*e2).j << ") ";

	h1.Release();

	++e2;

	cout << " [2.2]=ValueType(" << (*e2).i << ", " << (*e2).j << ") ";
	if (!++e2) cout << " ";
	cout << endl;
}

void ContainerHandledListExp()
{
	ContainerHandledListTypecasting();
	cout << endl;
	
	typedef HandledList<int> Reg1;
	
	Reg1 r1;
	
	{
		//   
		cout << "  : ";
		Reg1::Enumerator e11 = r1.GetEnumerator();
		while (e11.Next()) cout << e11.Current() << " !!!";
		cout << endl;
	}
	
	Reg1::Handle h11 = r1.Add(1);
	
	//    1 -
	cout << "   1 -: ";
	Reg1::Enumerator e12 = r1.GetEnumerator();
	while (e12.Next()) cout << e12.Current();
	cout << endl;
	
	Reg1::Handle h12 = r1.Add(2);
	
	//   
	cout << "  : ";
	while (e12.Next()) cout << e12.Current() << ", ";
	cout << endl;
	
	//    2 -
	cout << "   2 -: ";
	Reg1::Enumerator e13 = r1.GetEnumerator();
	while (++e13) cout << *e13 << ", ";
	cout << endl;
	
	//  2-  
	cout << " 2-  : ";
	e12.Reset(); e13.Reset();
	while (++e13 && ++e12) cout << *e12 << "=" << *e13 << ", ";
	cout << endl;
	
	//      
	cout << "     : ";
	e12.Reset();
	r1.Add(99); //  
	++e12; cout << "[1]=" << *e12;
	++e12; cout << ", [2]=" << *e12;

	Reg1::Enumerator e15 = e12;
	

	Reg1::Handle h13 = r1.Add(3);

	++e12; cout << ", [3]=" << *e12;
	
	Reg1::Handle h14 = h13;

	++e12; cout << ", [4]=" << *e12;
	
	h14 = h11; //   3   1

	++e12; cout << ", [4+]=" << *e12;
	if (!++e12) cout << " ";
	cout << endl;

	//  
	cout << " : ";
	cout << ", [2]=" << *e15;
	while (++e15) cout << *e15 << ", ";
	cout << endl;
	
	//    
	cout << "   : ";
	++e12; cout << "[1]=" << *e12;
	h11.Release();
	h12.Release();
	++e12; cout << ", [2]=" << *e12;
	++e12; cout << ", [3]=" << *e12;
	h13.Release();
	h14.Release();
	++e12;
	if (!e12) cout << " ";
	cout << endl;
	
}


void ContainerRegisterTypecasting()
{
	//     
	cerr << "    : ";
	typedef Register<ValueType, StoreType> Reg0;
	Reg0 r0;
	
	cerr << endl << "     ValueType(2, 5) ";
	Reg0::Item h1(r0, ValueType(2, 5));
	
	cerr << endl << "      ";
	Reg0::Item h2 = h1;	
	
	cerr << endl << "          ";
	Reg0::Enumerator e1 = r0.GetEnumerator();
	++e1;

	cerr << endl << "      ";
	Reg0::Enumerator e2 = e1;

	cerr << endl << "    ";
	cerr << " [1.1]=ValueType(" << e1->i << ", " << e1->j << ") ";
	++e1;
	cerr << " [1.2]=ValueType(" << e1->i << ", " << e1->j << ") ";

	cerr << endl << "    ";
	cerr << " [2.1]=ValueType(" << (*e2).i << ", " << (*e2).j << ") ";

	h1.Unassign();

	++e2;

	cerr << " [2.2]=ValueType(" << (*e2).i << ", " << (*e2).j << ") ";
	if (!++e2) cerr << " ";
	cerr << endl;
}

void ContainerRegisterExTypecasting()
{
	//     
	cerr << "    : ";
	typedef RegisterEx<ValueType, StoreType> Reg0;
	Reg0::Ptr r0(new Reg0);
	
	cerr << endl << "     ValueType(2, 5) ";
	Reg0::Item h1(r0, ValueType(2, 5));
	
	cerr << endl << "      ";
	Reg0::Item h2 = h1;	
	
	cerr << endl << "          ";
	Reg0::Enumerator e1 = r0->GetEnumerator();
	++e1;

	cerr << endl << "       ";
	r0.SetNull();
	
	cerr << endl << "      ";
	Reg0::Enumerator e2 = e1;

	cerr << endl << "    ";
	cerr << " [1.1]=ValueType(" << e1->i << ", " << e1->j << ") ";
	++e1;
	cerr << " [1.2]=ValueType(" << e1->i << ", " << e1->j << ") ";

	cerr << endl << "    ";
	cerr << " [2.1]=ValueType(" << (*e2).i << ", " << (*e2).j << ") ";

	h1.Unassign();

	++e2;

	cerr << " [2.2]=ValueType(" << (*e2).i << ", " << (*e2).j << ") ";
	if (!++e2) cerr << " ";
	cerr << endl;
}

void ContainerRegisterExpNoDynamic()
{
	typedef Register<int> Reg1;
	
	Reg1 r1;

	//   
	cerr << "  : ";
	Reg1::Enumerator e11 = r1.GetEnumerator(false);
	while (e11.Next()) cerr << e11.Current() << " !!!";
	cerr << endl;

	//  -
	Reg1::Item h11(r1, 1);

	//    
	cerr << "  : ";
	while (e11.Next()) cerr << e11.Current() << " !!!";
	cerr << endl;

	cerr << "1   : ";cerr << endl;
	//  2 -
	Reg1::Item h12(r1, 2);
	//  
	Reg1::Enumerator e12 = r1.GetEnumerator(false);
	//  3 -
	Reg1::Item h13(r1, 3);
	//  2 -
	e12.Next(); e12.Next(); if (!e12) cerr << " !!!"<< endl;
	//  2 -
	h12.Unassign();
	//   3 - - 
	e12.Next(); if (e12) cerr << " !!!"<< endl;
	//  1 -
	e12.Next(); if (e12.Current() != 1) cerr << " !!!" << endl;
	//  3 -
	h13.Unassign();
	//   - 
	e12.Next(); if (e12) cerr << " !!!"<< endl;
	//  1 -
	e12.Next(); if (!e12) cerr << " !!!"<< endl;
	//  1 -
	h11.Unassign();
	//  4 -
	Reg1::Item h14(r1, 4);
	//   - - 
	e12.Next(); if (e12) cerr << " !!!"<< endl;
	//   - - 
	e12.Next(); if (e12) cerr << " !!!"<< endl;

	cerr << "2   : ";cerr << endl;
	Reg1 r2;
	Reg1::Item h21(r2, 1);
	Reg1::Item h22(r2, 2);
	//  
	Reg1::Enumerator e21 = r2.GetEnumerator(false);
	//  1 -
	e21.Next(); if (e21.Current() != 1) cerr << " !!!" << endl;
	//   2
	Reg1::Enumerator e22 = e21;
	//  2 -
	e21.Next(); if (e21.Current() != 2) cerr << " !!!" << endl;
	//  2 -
	h22.Unassign();
	//  1 -
	h21.Unassign();
	//  3 -
	Reg1::Item h23(r2, 3);
	//   3
	Reg1::Enumerator e23 = e21;
	//  2  - 
	e22.Next(); if (e22) cerr << " !!!"<< endl;
	//  3  - 
	e23.Next(); if (e23) cerr << " !!!"<< endl;
	
	cerr << "3   : ";cerr << endl;
	Reg1 r3;
	Reg1::Item h31(r3, 1);
	Reg1::Item h32(r3, 2);
	//  
	Reg1::Enumerator e31 = r3.GetEnumerator(false);
	//  2 -
	e31.Next(); e31.Next(); if (!e31) cerr << " !!!"<< endl;
	//  1 -
	h31.Unassign();
	//  - 
	e31.Next(); if (e31) cerr << " !!!"<< endl;
	//  2 -
	e31.Next(); if (!e31) cerr << " !!!"<< endl;
	//  3 -
	Reg1::Item h33(r3, 3);
	//  - 
	e31.Next(); if (e31) cerr << " !!!"<< endl;
}

void ContainerRegisterExp()
{
	ContainerRegisterTypecasting();
	cerr << endl;
	ContainerRegisterExTypecasting();
	cerr << endl;
	ContainerRegisterExpNoDynamic();
	cerr << endl;
	
	typedef Register<int> Reg1;
	
	Reg1 r1;
	
	{
		//   
		cerr << "  : ";
		Reg1::Enumerator e11 = r1.GetEnumerator();
		while (e11.Next()) cerr << e11.Current() << " !!!";
		cerr << endl;
	}
	
	Reg1::Item h11(r1, 1);
	
	//    1 -
	cerr << "   1 -: ";
	Reg1::Enumerator e12 = r1.GetEnumerator();
	while (e12.Next()) cerr << e12.Current();
	cerr << endl;
	
	Reg1::Item h12(r1, 2);
	
	//   
	cerr << "  : ";
	while (e12.Next()) cerr << e12.Current() << ", ";
	cerr << endl;
	
	//    2 -
	cerr << "   2 -: ";
	Reg1::Enumerator e13 = r1.GetEnumerator();
	while (++e13) cerr << *e13 << ", ";
	cerr << endl;
	
	//  2-  
	cerr << " 2-  : ";
	e12.Reset(); e13.Reset();
	while (++e13 && ++e12) cerr << *e12 << "=" << *e13 << ", ";
	cerr << endl;
	
	//      
	cerr << "     : ";
	e12.Reset();
	++e12; cerr << "[1]=" << *e12;
	++e12; cerr << ", [2]=" << *e12;

	Reg1::Enumerator e15 = e12;
	

	Reg1::Item h13(r1, 3);

	++e12; cerr << ", [3]=" << *e12;
	
	Reg1::Item h14 = h13;

	++e12; cerr << ", [4]=" << *e12;
	
	h14 = h11; //   3   1

	++e12; cerr << ", [4+]=" << *e12;
	if (!++e12) cerr << " ";
	cerr << endl;

	//  
	cerr << " : ";
	cerr << ", [2]=" << *e15;
	while (++e15) cerr << *e15 << ", ";
	cerr << endl;
	
	//    
	cerr << "   : ";
	++e12; cerr << "[1]=" << *e12;
	h11.Unassign();
	h12.Unassign();
	++e12; cerr << ", [2]=" << *e12;
	++e12; cerr << ", [3]=" << *e12;
	h13.Unassign();
	h14.Unassign();
	++e12;
	if (!e12) cerr << " ";
	cerr << endl;
}
